# Stubs for compileall (Python 3)

import os
import sys
from typing import Optional, Union, Pattern

if sys.version_info < (3, 6):
    _Path = Union[str, bytes]
    _SuccessType = bool
else:
    _Path = Union[str, bytes, os.PathLike]
    _SuccessType = int

# rx can be any object with a 'search' method; once we have Protocols we can change the type
if sys.version_info < (3, 5):
    def compile_dir(dir: _Path, maxlevels: int = ..., ddir: Optional[_Path] = ..., force: bool = ..., rx: Optional[Pattern] = ..., quiet: int = ..., legacy: bool = ..., optimize: int = ...) -> _SuccessType: ...
else:
    def compile_dir(dir: _Path, maxlevels: int = ..., ddir: Optional[_Path] = ..., force: bool = ..., rx: Optional[Pattern] = ..., quiet: int = ..., legacy: bool = ..., optimize: int = ..., workers: int = ...) -> _SuccessType: ...
def compile_file(fullname: _Path, ddir: Optional[_Path] = ..., force: bool = ..., rx: Optional[Pattern] = ..., quiet: int = ..., legacy: bool = ..., optimize: int = ...) -> _SuccessType: ...
def compile_path(skip_curdir: bool = ..., maxlevels: int = ..., force: bool = ..., quiet: int = ..., legacy: bool = ..., optimize: int = ...) -> _SuccessType: ...



import random
import copy

from ... import colors
from ... import console
from ... import shapes
from ... import funcs


class Sprite:

    def __init__(self, image, position=(0, 0), color=None, char=None):
        """
            image = Image() instance
            position = (int x, int y)
            color = int range(0, 8)
            char = str char
        """

        self.image = copy.deepcopy(image)
        self.position = [int(position[0]), int(position[1])]

        self.color = color if color else random.randint(1, 8)
        self._char = str(char)[0] if char else console.supportedChars('●', 'O')

    def __repr__(self):
        return ('Sprite({!r}, position={!r}, color={!r}, char={!r})'
                .format(self.image, self.position, self.color, self._char))

    def char(self, pos):
        try:
            char = self.image.char(pos)
            if char is True or char is False:
                raise AttributeError
        except AttributeError:
            char = self._char
        return char

    def move(self, direction=0):
        """
            direction = int range(0, 4)

            0 = Down
            1 = Left
            2 = Up
            3 = Right
        """
        direction = int(direction)
        self.position = list(self.position)

        if direction == 0:
            self.position[1] += 1
        elif direction == 1:
            self.position[0] -= 1
        elif direction == 2:
            self.position[1] -= 1
        elif direction == 3:
            self.position[0] += 1

    def touching(self, canvas, side=None):
        """
            Returns True if touching any pixels [on specified side].

            0 = Bottom
            1 = Left
            2 = Top
            3 = Right
            None = All

            This works by creating a list of the image rotated so all the
                requested sides are facing up, then it finds the top edge for
                each image and rotates the coordinates back to the original
                image.
        """
        try:
            sides = list(side)
        except TypeError:
            if side is None:
                sides = list(range(4))
            else:
                sides = [side]

        # Generate rotated images for each direction.
        images = {}
        image = self.image.image()
        for side in sides:
            images.update({
                (side + 2) % 4:
                    funcs.rotateImage(image, (side + 2) % 4)
            })

        # Go through each image finding top edge,
        #   then rotate coordinates to match original image.
        edges = []
        for side, image in images.items():

            for x in range(len(image[0])):
                y = 0
                # If the first pixel is True, no look any further.
                if not image[0][x]:
                    while not image[y][x]:
                        y += 1
                        if y >= len(image):
                            # Fallen off bottom of image, therefore no edge.
                            y = None
                            break

                # Don't do anything if no pixels in column.
                if not y is None:
                    y -= 1  # To get pixel next to pixel which is on.

                    # Get coordinates the right way around.
                    pos = (x, y)
                    size = [len(image), len(image[0])]
                    for i in range(4 - side):
                        size.reverse()
                        pos = funcs.rotatePosition(pos, size)
                    edges.append(pos)

        # Find if any other sprites are in our edge coordinates.
        for pixel in edges:
            pixel = (int(self.position[0] + pixel[0]),
                     int(self.position[1] + pixel[1]))
            if canvas.testPixel(pixel):
                return True
        return False

    def overlaps(self, canvas, exclude=[]):
        """
            Returns True if sprite is touching any other sprite.
        """
        try:
            exclude = list(exclude)
        except TypeError:
            exclude = [exclude]
        exclude.append(self)

        for selfY, row in enumerate(self.image.image()):
            for selfX, pixel in enumerate(row):
                canvasPixelOn = canvas.testPixel(
                    (selfX + self.position[0], selfY + self.position[1]),
                    excludedSprites=exclude
                )
                if pixel and canvasPixelOn:
                    return True
        return False

    def onEdge(self, canvas):
        """
            Returns a list of the sides of the sprite
                which are touching the edge of the canvas.

            0 = Bottom
            1 = Left
            2 = Top
            3 = Right
        """
        sides = []
        if int(self.position[0]) <= 0:
            sides.append(1)

        if (int(self.position[0]) + self.image.width) >= canvas.width:
            sides.append(3)

        if int(self.position[1]) <= 0:
            sides.append(2)

        if (int(self.position[1]) + self.image.height) >= canvas.height:
            sides.append(0)

        return sides



# Stubs for distutils.cmd

from typing import Callable, List, Tuple, Union, Optional, Iterable, Any, Text
from abc import abstractmethod
from distutils.dist import Distribution

class Command:
    sub_commands: List[Tuple[str, Union[Callable[[], bool], str, None]]]
    def __init__(self, dist: Distribution) -> None: ...
    @abstractmethod
    def initialize_options(self) -> None: ...
    @abstractmethod
    def finalize_options(self) -> None: ...
    @abstractmethod
    def run(self) -> None: ...

    def announce(self, msg: Text, level: int = ...) -> None: ...
    def debug_print(self, msg: Text) -> None: ...

    def ensure_string(self, option: str, default: Optional[str] = ...) -> None: ...
    def ensure_string_list(self, option: Union[str, List[str]]) -> None: ...
    def ensure_filename(self, option: str) -> None: ...
    def ensure_dirname(self, option: str) -> None: ...

    def get_command_name(self) -> str: ...
    def set_undefined_options(self, src_cmd: Text, *option_pairs: Tuple[str, str]) -> None: ...
    def get_finalized_command(self, command: Text, create: int = ...) -> Command: ...
    def reinitialize_command(self, command: Union[Command, Text], reinit_subcommands: int = ...) -> Command: ...
    def run_command(self, command: Text) -> None: ...
    def get_sub_commands(self) -> List[str]: ...

    def warn(self, msg: Text) -> None: ...
    def execute(self, func: Callable[..., Any], args: Iterable[Any], msg: Optional[Text] = ..., level: int = ...) -> None: ...
    def mkpath(self, name: str, mode: int = ...) -> None: ...
    def copy_file(self, infile: str, outfile: str, preserve_mode: int = ..., preserve_times: int = ..., link: Optional[str] = ..., level: Any = ...) -> Tuple[str, bool]: ...  # level is not used
    def copy_tree(self, infile: str, outfile: str, preserve_mode: int = ..., preserve_times: int = ..., preserve_symlinks: int = ..., level: Any = ...) -> List[str]: ...  # level is not used
    def move_file(self, src: str, dest: str, level: Any = ...) -> str: ...  # level is not used
    def spawn(self, cmd: Iterable[str], search_path: int = ..., level: Any = ...) -> None: ...  # level is not used
    def make_archive(self, base_name: str, format: str, root_dir: Optional[str] = ..., base_dir: Optional[str] = ..., owner: Optional[str] = ..., group: Optional[str] = ...) -> str: ...
    def make_file(self, infiles: Union[str, List[str], Tuple[str]], outfile: str, func: Callable[..., Any], args: List[Any], exec_msg: Optional[str] = ..., skip_msg: Optional[str] = ..., level: Any = ...) -> None: ...  # level is not used



# Generated from JSONPath.g4 by ANTLR 4.8
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\26")
        buf.write("\u00a1\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\3\2\3\2\7\2%\n\2\f")
        buf.write("\2\16\2(\13\2\3\2\3\2\3\3\3\3\5\3.\n\3\3\4\3\4\3\4\3\4")
        buf.write("\5\4\64\n\4\3\5\3\5\3\5\3\5\5\5:\n\5\3\5\3\5\5\5>\n\5")
        buf.write("\3\5\3\5\5\5B\n\5\3\5\3\5\5\5F\n\5\3\5\5\5I\n\5\3\6\3")
        buf.write("\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6")
        buf.write("\3\6\3\6\3\6\3\6\3\6\3\6\5\6`\n\6\3\7\3\7\3\7\3\7\3\7")
        buf.write("\3\b\3\b\3\b\3\b\3\b\3\t\3\t\3\t\3\n\3\n\3\n\3\n\6\ns")
        buf.write("\n\n\r\n\16\nt\3\n\3\n\3\13\3\13\5\13{\n\13\3\13\3\13")
        buf.write("\5\13\177\n\13\3\13\3\13\5\13\u0083\n\13\3\13\3\13\3\f")
        buf.write("\3\f\3\f\5\f\u008a\n\f\3\f\3\f\5\f\u008e\n\f\3\f\5\f\u0091")
        buf.write("\n\f\3\r\3\r\3\16\3\16\3\16\3\16\3\17\3\17\3\17\3\17\3")
        buf.write("\20\3\20\3\21\3\21\3\21\2\2\22\2\4\6\b\n\f\16\20\22\24")
        buf.write("\26\30\32\34\36 \2\4\3\2\b\n\3\2\23\24\2\u00a6\2\"\3\2")
        buf.write("\2\2\4-\3\2\2\2\6\63\3\2\2\2\bH\3\2\2\2\n_\3\2\2\2\fa")
        buf.write("\3\2\2\2\16f\3\2\2\2\20k\3\2\2\2\22n\3\2\2\2\24x\3\2\2")
        buf.write("\2\26\u0090\3\2\2\2\30\u0092\3\2\2\2\32\u0094\3\2\2\2")
        buf.write("\34\u0098\3\2\2\2\36\u009c\3\2\2\2 \u009e\3\2\2\2\"&\7")
        buf.write("\17\2\2#%\5\4\3\2$#\3\2\2\2%(\3\2\2\2&$\3\2\2\2&\'\3\2")
        buf.write("\2\2\')\3\2\2\2(&\3\2\2\2)*\7\2\2\3*\3\3\2\2\2+.\5\b\5")
        buf.write("\2,.\5\6\4\2-+\3\2\2\2-,\3\2\2\2.\5\3\2\2\2/\60\7\22\2")
        buf.write("\2\60\64\5\30\r\2\61\62\7\22\2\2\62\64\5\32\16\2\63/\3")
        buf.write("\2\2\2\63\61\3\2\2\2\64\7\3\2\2\2\65I\5\26\f\2\66\67\7")
        buf.write("\21\2\2\67I\5\36\20\28:\7\21\2\298\3\2\2\29:\3\2\2\2:")
        buf.write(";\3\2\2\2;I\5 \21\2<>\7\21\2\2=<\3\2\2\2=>\3\2\2\2>?\3")
        buf.write("\2\2\2?I\5\n\6\2@B\7\21\2\2A@\3\2\2\2AB\3\2\2\2BC\3\2")
        buf.write("\2\2CI\5\24\13\2DF\7\21\2\2ED\3\2\2\2EF\3\2\2\2FG\3\2")
        buf.write("\2\2GI\5\22\n\2H\65\3\2\2\2H\66\3\2\2\2H9\3\2\2\2H=\3")
        buf.write("\2\2\2HA\3\2\2\2HE\3\2\2\2I\t\3\2\2\2JK\7\3\2\2KL\7\4")
        buf.write("\2\2LM\7\5\2\2MN\5\f\7\2NO\7\6\2\2OP\7\7\2\2P`\3\2\2\2")
        buf.write("QR\7\3\2\2RS\7\4\2\2ST\7\5\2\2TU\5\16\b\2UV\7\6\2\2VW")
        buf.write("\7\7\2\2W`\3\2\2\2XY\7\3\2\2YZ\7\4\2\2Z[\7\5\2\2[\\\5")
        buf.write("\20\t\2\\]\7\6\2\2]^\7\7\2\2^`\3\2\2\2_J\3\2\2\2_Q\3\2")
        buf.write("\2\2_X\3\2\2\2`\13\3\2\2\2ab\7\20\2\2bc\5\26\f\2cd\t\2")
        buf.write("\2\2de\7\25\2\2e\r\3\2\2\2fg\7\20\2\2gh\5\26\f\2hi\7\n")
        buf.write("\2\2ij\7\24\2\2j\17\3\2\2\2kl\7\20\2\2lm\5\26\f\2m\21")
        buf.write("\3\2\2\2no\7\3\2\2or\t\3\2\2pq\7\13\2\2qs\t\3\2\2rp\3")
        buf.write("\2\2\2st\3\2\2\2tr\3\2\2\2tu\3\2\2\2uv\3\2\2\2vw\7\7\2")
        buf.write("\2w\23\3\2\2\2xz\7\3\2\2y{\7\25\2\2zy\3\2\2\2z{\3\2\2")
        buf.write("\2{|\3\2\2\2|~\7\f\2\2}\177\7\25\2\2~}\3\2\2\2~\177\3")
        buf.write("\2\2\2\177\u0082\3\2\2\2\u0080\u0081\7\f\2\2\u0081\u0083")
        buf.write("\7\25\2\2\u0082\u0080\3\2\2\2\u0082\u0083\3\2\2\2\u0083")
        buf.write("\u0084\3\2\2\2\u0084\u0085\7\7\2\2\u0085\25\3\2\2\2\u0086")
        buf.write("\u0087\7\21\2\2\u0087\u0091\5\30\r\2\u0088\u008a\7\21")
        buf.write("\2\2\u0089\u0088\3\2\2\2\u0089\u008a\3\2\2\2\u008a\u008b")
        buf.write("\3\2\2\2\u008b\u0091\5\32\16\2\u008c\u008e\7\21\2\2\u008d")
        buf.write("\u008c\3\2\2\2\u008d\u008e\3\2\2\2\u008e\u008f\3\2\2\2")
        buf.write("\u008f\u0091\5\34\17\2\u0090\u0086\3\2\2\2\u0090\u0089")
        buf.write("\3\2\2\2\u0090\u008d\3\2\2\2\u0091\27\3\2\2\2\u0092\u0093")
        buf.write("\7\23\2\2\u0093\31\3\2\2\2\u0094\u0095\7\3\2\2\u0095\u0096")
        buf.write("\t\3\2\2\u0096\u0097\7\7\2\2\u0097\33\3\2\2\2\u0098\u0099")
        buf.write("\7\3\2\2\u0099\u009a\7\25\2\2\u009a\u009b\7\7\2\2\u009b")
        buf.write("\35\3\2\2\2\u009c\u009d\7\r\2\2\u009d\37\3\2\2\2\u009e")
        buf.write("\u009f\7\16\2\2\u009f!\3\2\2\2\22&-\639=AEH_tz~\u0082")
        buf.write("\u0089\u008d\u0090")
        return buf.getvalue()


class JSONPathParser(Parser):
    grammarFileName = "JSONPath.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = ["<INVALID>", "'['", "'?'", "'('", "')'", "']'", "'>'",
                    "'<'", "'=='", "','", "':'", "'*'", "'[*]'", "'$'",
                    "'@'", "'.'", "'..'"]

    symbolicNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "ROOT",
        "CURRENT",
        "SINGLE_DOT",
        "DOUBLE_DOT",
        "LETTER",
        "STRING",
        "INT",
        "WS"]

    RULE_jsonpath = 0
    RULE_expression = 1
    RULE_doubleDotExpression = 2
    RULE_singleDotExpression = 3
    RULE_filters = 4
    RULE_numericFilter = 5
    RULE_stringFilter = 6
    RULE_booleanFilter = 7
    RULE_union = 8
    RULE_arraySlice = 9
    RULE_fieldAccessor = 10
    RULE_field = 11
    RULE_bracketField = 12
    RULE_arrayIndex = 13
    RULE_wildcard = 14
    RULE_bracketWildcard = 15

    ruleNames = [
        "jsonpath",
        "expression",
        "doubleDotExpression",
        "singleDotExpression",
        "filters",
        "numericFilter",
        "stringFilter",
        "booleanFilter",
        "union",
        "arraySlice",
        "fieldAccessor",
        "field",
        "bracketField",
        "arrayIndex",
        "wildcard",
        "bracketWildcard"]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    T__4 = 5
    T__5 = 6
    T__6 = 7
    T__7 = 8
    T__8 = 9
    T__9 = 10
    T__10 = 11
    T__11 = 12
    ROOT = 13
    CURRENT = 14
    SINGLE_DOT = 15
    DOUBLE_DOT = 16
    LETTER = 17
    STRING = 18
    INT = 19
    WS = 20

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.8")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None

    class JsonpathContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ROOT(self):
            return self.getToken(JSONPathParser.ROOT, 0)

        def EOF(self):
            return self.getToken(JSONPathParser.EOF, 0)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    JSONPathParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(
                    JSONPathParser.ExpressionContext, i)

        def getRuleIndex(self):
            return JSONPathParser.RULE_jsonpath

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJsonpath"):
                listener.enterJsonpath(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJsonpath"):
                listener.exitJsonpath(self)

    def jsonpath(self):

        localctx = JSONPathParser.JsonpathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_jsonpath)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 32
            self.match(JSONPathParser.ROOT)
            self.state = 36
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (
                (
                    (_la) & ~0x3f) == 0 and (
                    (1 << _la) & (
                        (1 << JSONPathParser.T__0) | (
                    1 << JSONPathParser.T__11) | (
                        1 << JSONPathParser.SINGLE_DOT) | (
                    1 << JSONPathParser.DOUBLE_DOT))) != 0):
                self.state = 33
                self.expression()
                self.state = 38
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 39
            self.match(JSONPathParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def singleDotExpression(self):
            return self.getTypedRuleContext(
                JSONPathParser.SingleDotExpressionContext, 0)

        def doubleDotExpression(self):
            return self.getTypedRuleContext(
                JSONPathParser.DoubleDotExpressionContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpression"):
                listener.enterExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpression"):
                listener.exitExpression(self)

    def expression(self):

        localctx = JSONPathParser.ExpressionContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_expression)
        try:
            self.state = 43
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [JSONPathParser.T__0, JSONPathParser.T__11,
                         JSONPathParser.SINGLE_DOT]:
                self.enterOuterAlt(localctx, 1)
                self.state = 41
                self.singleDotExpression()
                pass
            elif token in [JSONPathParser.DOUBLE_DOT]:
                self.enterOuterAlt(localctx, 2)
                self.state = 42
                self.doubleDotExpression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DoubleDotExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOUBLE_DOT(self):
            return self.getToken(JSONPathParser.DOUBLE_DOT, 0)

        def field(self):
            return self.getTypedRuleContext(JSONPathParser.FieldContext, 0)

        def bracketField(self):
            return self.getTypedRuleContext(
                JSONPathParser.BracketFieldContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_doubleDotExpression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDoubleDotExpression"):
                listener.enterDoubleDotExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDoubleDotExpression"):
                listener.exitDoubleDotExpression(self)

    def doubleDotExpression(self):

        localctx = JSONPathParser.DoubleDotExpressionContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_doubleDotExpression)
        try:
            self.state = 49
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 45
                self.match(JSONPathParser.DOUBLE_DOT)
                self.state = 46
                self.field()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 47
                self.match(JSONPathParser.DOUBLE_DOT)
                self.state = 48
                self.bracketField()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SingleDotExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def fieldAccessor(self):
            return self.getTypedRuleContext(
                JSONPathParser.FieldAccessorContext, 0)

        def SINGLE_DOT(self):
            return self.getToken(JSONPathParser.SINGLE_DOT, 0)

        def wildcard(self):
            return self.getTypedRuleContext(JSONPathParser.WildcardContext, 0)

        def bracketWildcard(self):
            return self.getTypedRuleContext(
                JSONPathParser.BracketWildcardContext, 0)

        def filters(self):
            return self.getTypedRuleContext(JSONPathParser.FiltersContext, 0)

        def arraySlice(self):
            return self.getTypedRuleContext(
                JSONPathParser.ArraySliceContext, 0)

        def union(self):
            return self.getTypedRuleContext(JSONPathParser.UnionContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_singleDotExpression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSingleDotExpression"):
                listener.enterSingleDotExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSingleDotExpression"):
                listener.exitSingleDotExpression(self)

    def singleDotExpression(self):

        localctx = JSONPathParser.SingleDotExpressionContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_singleDotExpression)
        self._la = 0  # Token type
        try:
            self.state = 70
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 7, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 51
                self.fieldAccessor()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 52
                self.match(JSONPathParser.SINGLE_DOT)
                self.state = 53
                self.wildcard()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 55
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 54
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 57
                self.bracketWildcard()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 59
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 58
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 61
                self.filters()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 63
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 62
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 65
                self.arraySlice()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 67
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 66
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 69
                self.union()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FiltersContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def numericFilter(self):
            return self.getTypedRuleContext(
                JSONPathParser.NumericFilterContext, 0)

        def stringFilter(self):
            return self.getTypedRuleContext(
                JSONPathParser.StringFilterContext, 0)

        def booleanFilter(self):
            return self.getTypedRuleContext(
                JSONPathParser.BooleanFilterContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_filters

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFilters"):
                listener.enterFilters(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFilters"):
                listener.exitFilters(self)

    def filters(self):

        localctx = JSONPathParser.FiltersContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_filters)
        try:
            self.state = 93
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 8, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 72
                self.match(JSONPathParser.T__0)
                self.state = 73
                self.match(JSONPathParser.T__1)
                self.state = 74
                self.match(JSONPathParser.T__2)
                self.state = 75
                self.numericFilter()
                self.state = 76
                self.match(JSONPathParser.T__3)
                self.state = 77
                self.match(JSONPathParser.T__4)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 79
                self.match(JSONPathParser.T__0)
                self.state = 80
                self.match(JSONPathParser.T__1)
                self.state = 81
                self.match(JSONPathParser.T__2)
                self.state = 82
                self.stringFilter()
                self.state = 83
                self.match(JSONPathParser.T__3)
                self.state = 84
                self.match(JSONPathParser.T__4)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 86
                self.match(JSONPathParser.T__0)
                self.state = 87
                self.match(JSONPathParser.T__1)
                self.state = 88
                self.match(JSONPathParser.T__2)
                self.state = 89
                self.booleanFilter()
                self.state = 90
                self.match(JSONPathParser.T__3)
                self.state = 91
                self.match(JSONPathParser.T__4)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumericFilterContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CURRENT(self):
            return self.getToken(JSONPathParser.CURRENT, 0)

        def fieldAccessor(self):
            return self.getTypedRuleContext(
                JSONPathParser.FieldAccessorContext, 0)

        def INT(self):
            return self.getToken(JSONPathParser.INT, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_numericFilter

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNumericFilter"):
                listener.enterNumericFilter(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNumericFilter"):
                listener.exitNumericFilter(self)

    def numericFilter(self):

        localctx = JSONPathParser.NumericFilterContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_numericFilter)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95
            self.match(JSONPathParser.CURRENT)
            self.state = 96
            self.fieldAccessor()
            self.state = 97
            _la = self._input.LA(1)
            if not (
                (((_la) & ~0x3f) == 0 and (
                    (1 << _la) & (
                        (1 << JSONPathParser.T__5) | (
                            1 << JSONPathParser.T__6) | (
                    1 << JSONPathParser.T__7))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 98
            self.match(JSONPathParser.INT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StringFilterContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CURRENT(self):
            return self.getToken(JSONPathParser.CURRENT, 0)

        def fieldAccessor(self):
            return self.getTypedRuleContext(
                JSONPathParser.FieldAccessorContext, 0)

        def STRING(self):
            return self.getToken(JSONPathParser.STRING, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_stringFilter

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStringFilter"):
                listener.enterStringFilter(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStringFilter"):
                listener.exitStringFilter(self)

    def stringFilter(self):

        localctx = JSONPathParser.StringFilterContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_stringFilter)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 100
            self.match(JSONPathParser.CURRENT)
            self.state = 101
            self.fieldAccessor()
            self.state = 102
            self.match(JSONPathParser.T__7)
            self.state = 103
            self.match(JSONPathParser.STRING)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class BooleanFilterContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CURRENT(self):
            return self.getToken(JSONPathParser.CURRENT, 0)

        def fieldAccessor(self):
            return self.getTypedRuleContext(
                JSONPathParser.FieldAccessorContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_booleanFilter

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBooleanFilter"):
                listener.enterBooleanFilter(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBooleanFilter"):
                listener.exitBooleanFilter(self)

    def booleanFilter(self):

        localctx = JSONPathParser.BooleanFilterContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_booleanFilter)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 105
            self.match(JSONPathParser.CURRENT)
            self.state = 106
            self.fieldAccessor()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class UnionContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self, i: int = None):
            if i is None:
                return self.getTokens(JSONPathParser.STRING)
            else:
                return self.getToken(JSONPathParser.STRING, i)

        def LETTER(self, i: int = None):
            if i is None:
                return self.getTokens(JSONPathParser.LETTER)
            else:
                return self.getToken(JSONPathParser.LETTER, i)

        def getRuleIndex(self):
            return JSONPathParser.RULE_union

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterUnion"):
                listener.enterUnion(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitUnion"):
                listener.exitUnion(self)

    def union(self):

        localctx = JSONPathParser.UnionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_union)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 108
            self.match(JSONPathParser.T__0)
            self.state = 109
            _la = self._input.LA(1)
            if not (_la == JSONPathParser.LETTER or _la ==
                    JSONPathParser.STRING):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 112
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 110
                self.match(JSONPathParser.T__8)
                self.state = 111
                _la = self._input.LA(1)
                if not (_la == JSONPathParser.LETTER or _la ==
                        JSONPathParser.STRING):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 114
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == JSONPathParser.T__8):
                    break

            self.state = 116
            self.match(JSONPathParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArraySliceContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self, i: int = None):
            if i is None:
                return self.getTokens(JSONPathParser.INT)
            else:
                return self.getToken(JSONPathParser.INT, i)

        def getRuleIndex(self):
            return JSONPathParser.RULE_arraySlice

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArraySlice"):
                listener.enterArraySlice(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArraySlice"):
                listener.exitArraySlice(self)

    def arraySlice(self):

        localctx = JSONPathParser.ArraySliceContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_arraySlice)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 118
            self.match(JSONPathParser.T__0)
            self.state = 120
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == JSONPathParser.INT:
                self.state = 119
                self.match(JSONPathParser.INT)

            self.state = 122
            self.match(JSONPathParser.T__9)
            self.state = 124
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == JSONPathParser.INT:
                self.state = 123
                self.match(JSONPathParser.INT)

            self.state = 128
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == JSONPathParser.T__9:
                self.state = 126
                self.match(JSONPathParser.T__9)
                self.state = 127
                self.match(JSONPathParser.INT)

            self.state = 130
            self.match(JSONPathParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldAccessorContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SINGLE_DOT(self):
            return self.getToken(JSONPathParser.SINGLE_DOT, 0)

        def field(self):
            return self.getTypedRuleContext(JSONPathParser.FieldContext, 0)

        def bracketField(self):
            return self.getTypedRuleContext(
                JSONPathParser.BracketFieldContext, 0)

        def arrayIndex(self):
            return self.getTypedRuleContext(
                JSONPathParser.ArrayIndexContext, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_fieldAccessor

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFieldAccessor"):
                listener.enterFieldAccessor(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFieldAccessor"):
                listener.exitFieldAccessor(self)

    def fieldAccessor(self):

        localctx = JSONPathParser.FieldAccessorContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_fieldAccessor)
        self._la = 0  # Token type
        try:
            self.state = 142
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 15, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 132
                self.match(JSONPathParser.SINGLE_DOT)
                self.state = 133
                self.field()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 135
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 134
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 137
                self.bracketField()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 139
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == JSONPathParser.SINGLE_DOT:
                    self.state = 138
                    self.match(JSONPathParser.SINGLE_DOT)

                self.state = 141
                self.arrayIndex()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LETTER(self):
            return self.getToken(JSONPathParser.LETTER, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_field

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterField"):
                listener.enterField(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitField"):
                listener.exitField(self)

    def field(self):

        localctx = JSONPathParser.FieldContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_field)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 144
            self.match(JSONPathParser.LETTER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class BracketFieldContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JSONPathParser.STRING, 0)

        def LETTER(self):
            return self.getToken(JSONPathParser.LETTER, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_bracketField

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBracketField"):
                listener.enterBracketField(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBracketField"):
                listener.exitBracketField(self)

    def bracketField(self):

        localctx = JSONPathParser.BracketFieldContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_bracketField)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 146
            self.match(JSONPathParser.T__0)
            self.state = 147
            _la = self._input.LA(1)
            if not (_la == JSONPathParser.LETTER or _la ==
                    JSONPathParser.STRING):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 148
            self.match(JSONPathParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArrayIndexContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(JSONPathParser.INT, 0)

        def getRuleIndex(self):
            return JSONPathParser.RULE_arrayIndex

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArrayIndex"):
                listener.enterArrayIndex(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArrayIndex"):
                listener.exitArrayIndex(self)

    def arrayIndex(self):

        localctx = JSONPathParser.ArrayIndexContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_arrayIndex)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 150
            self.match(JSONPathParser.T__0)
            self.state = 151
            self.match(JSONPathParser.INT)
            self.state = 152
            self.match(JSONPathParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class WildcardContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return JSONPathParser.RULE_wildcard

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWildcard"):
                listener.enterWildcard(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWildcard"):
                listener.exitWildcard(self)

    def wildcard(self):

        localctx = JSONPathParser.WildcardContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_wildcard)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 154
            self.match(JSONPathParser.T__10)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class BracketWildcardContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None,
                     invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return JSONPathParser.RULE_bracketWildcard

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBracketWildcard"):
                listener.enterBracketWildcard(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBracketWildcard"):
                listener.exitBracketWildcard(self)

    def bracketWildcard(self):

        localctx = JSONPathParser.BracketWildcardContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_bracketWildcard)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 156
            self.match(JSONPathParser.T__11)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



import argparse

from platform import system

from . import meta


def parse_args():
  parser = argparse.ArgumentParser(
    meta.NAME,
    description=meta.DESC)

  parser.add_argument(
    '-V', '--version',
    help='print version and exit',
    action='store_true')

  parser.add_argument(
    'filename',
    nargs='?',
    type=str,
    help='filename of an image')

  parser.add_argument(
    '-c', '--charset',
    type=str,
    help='''character set to be used in ASCII mode (type `-c list` to list
    available charsets, the default is simple)''',
    default='simple'
  )

  parser.add_argument(
    '-i', '--invert-background',
    help='invert grayscale in ASCII mode',
    action='store_true')

  parser.add_argument(
    '-m', '--method',
    type=str,
    help='''name of a rendering method (use `-m list` to list available methods,
    the default is a24h)''',
    metavar='METHOD',
    default='a24h' if system() != "Windows" else "ascii")

  parser.add_argument(
    '-r', '--reduce-colors',
    type=int,
    help='reduce color palette of an input image (1-256)',
    metavar='N')

  parser.add_argument(
    '-s', '--size',
    type=int,
    help='width of an image',
    metavar='W')

  args = parser.parse_args()
  return parser, args



from typing import Any, Callable, Iterator, Optional, Tuple

class JSONEncoder:
    item_separator: str
    key_separator: str

    skipkeys: bool
    ensure_ascii: bool
    check_circular: bool
    allow_nan: bool
    sort_keys: bool
    indent: int

    def __init__(self, skipkeys: bool = ..., ensure_ascii: bool = ...,
                 check_circular: bool = ..., allow_nan: bool = ..., sort_keys: bool = ...,
                 indent: Optional[int] = ..., separators: Optional[Tuple[str, str]] = ...,
                 default: Optional[Callable] = ...) -> None: ...

    def default(self, o: Any) -> Any: ...
    def encode(self, o: Any) -> str: ...
    def iterencode(self, o: Any, _one_shot: bool = ...) -> Iterator[str]: ...



# Stubs for six.moves.urllib.response
from urllib import addbase as addbase
from urllib import addclosehook as addclosehook
from urllib import addinfo as addinfo
from urllib import addinfourl as addinfourl



# Stubs for http.server (Python 3.4)

import sys
from typing import Any, BinaryIO, Dict, List, Mapping, Optional, Tuple, Union
import socketserver
import email.message

if sys.version_info >= (3, 7):
    from builtins import _PathLike

class HTTPServer(socketserver.TCPServer):
    server_name: str
    server_port: int
    def __init__(self, server_address: Tuple[str, int],
                 RequestHandlerClass: type) -> None: ...

class BaseHTTPRequestHandler:
    client_address: Tuple[str, int]
    server: socketserver.BaseServer
    close_connection: bool
    requestline: str
    command: str
    path: str
    request_version: str
    headers: email.message.Message
    rfile: BinaryIO
    wfile: BinaryIO
    server_version: str
    sys_version: str
    error_message_format: str
    error_content_type: str
    protocol_version: str
    MessageClass: type
    responses: Mapping[int, Tuple[str, str]]
    def __init__(self, request: bytes, client_address: Tuple[str, int],
                 server: socketserver.BaseServer) -> None: ...
    def handle(self) -> None: ...
    def handle_one_request(self) -> None: ...
    def handle_expect_100(self) -> bool: ...
    def send_error(self, code: int, message: Optional[str] = ...,
                   explain: Optional[str] = ...) -> None: ...
    def send_response(self, code: int,
                      message: Optional[str] = ...) -> None: ...
    def send_header(self, keyword: str, value: str) -> None: ...
    def send_response_only(self, code: int,
                           message: Optional[str] = ...) -> None: ...
    def end_headers(self) -> None: ...
    def flush_headers(self) -> None: ...
    def log_request(self, code: Union[int, str] = ...,
                    size: Union[int, str] = ...) -> None: ...
    def log_error(self, format: str, *args: Any) -> None: ...
    def log_message(self, format: str, *args: Any) -> None: ...
    def version_string(self) -> str: ...
    def date_time_string(self, timestamp: Optional[int] = ...) -> str: ...
    def log_date_time_string(self) -> str: ...
    def address_string(self) -> str: ...

class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
    extensions_map: Dict[str, str]
    if sys.version_info >= (3, 7):
        def __init__(self, request: bytes, client_address: Tuple[str, int],
                     server: socketserver.BaseServer, directory: Optional[Union[str, _PathLike[str]]]) -> None: ...
    else:
        def __init__(self, request: bytes, client_address: Tuple[str, int],
                     server: socketserver.BaseServer) -> None: ...
    def do_GET(self) -> None: ...
    def do_HEAD(self) -> None: ...

class CGIHTTPRequestHandler(SimpleHTTPRequestHandler):
    cgi_directories: List[str]
    def do_POST(self) -> None: ...



from overrides import overrides

from ..json_widget import JSONWidget


class ArrayEndWidget(JSONWidget):
    """
    display widget for JSON `array` type node
    """

    def __init__(self, node):
        # display_key is not important for end widget
        super().__init__(node, True, False)

    @overrides
    def load_value_markup(self):
        return "]"



from typing import Optional

from .connection import S3Connection

from boto.connection import AWSAuthConnection
from boto.regioninfo import RegionInfo

from typing import List, Type, Text

class S3RegionInfo(RegionInfo):
    def connect(self, name: Optional[Text] = ..., endpoint: Optional[str] = ..., connection_cls: Optional[Type[AWSAuthConnection]] = ..., **kw_params) -> S3Connection: ...

def regions() -> List[S3RegionInfo]: ...
def connect_to_region(region_name: Text, **kw_params): ...



from typing import Dict, Union
from urllib.response import addinfourl

# Stubs for urllib.error

class URLError(IOError):
    reason: Union[str, BaseException]
class HTTPError(URLError, addinfourl):
    code: int
    headers: Dict[str, str]
    def __init__(self, url, code, msg, hdrs, fp) -> None: ...
class ContentTooShortError(URLError): ...
